@@ -1,3 +1,5 @@
+"""Run third-party tool (e.g. :code:`mypy`) against notebook or directory."""
+
 import argparse
 import configparser
 import os
@@ -6,7 +8,7 @@ import subprocess
 import sys
 import tempfile
 from pathlib import Path
-from typing import List
+from typing import Dict, Iterator, List, Optional, Tuple
 
 from nbqa import (
     __version__,
@@ -17,9 +19,29 @@ from nbqa import (
 )
 
 
-def _parse_args(raw_args):
+def _parse_args(raw_args: Optional[List[str]]) -> Tuple[str, str, List[str]]:
     """
     Parse command-line arguments.
+
+    Parameters
+    ----------
+    raw_args
+        Passed via command-line.
+
+    Returns
+    -------
+    command
+        The third-party tool to run (e.g. :code:`mypy`).
+    root_dir
+        The notebook or directory to run third-party tool on.
+    kwargs
+        Any additional flags passed to third-party tool (e.g. :code:`--quiet`).
+
+    Raises
+    ------
+    ValueError
+        If user doesn't specify both a command and a notebook/directory to run it
+        on (e.g. if the user runs :code:`nbqa flake8` instead of :code:`nbqa flake8 .`).
     """
     parser = argparse.ArgumentParser(
         description="Adapter to run any code-quality tool on a Jupyter notebook.",
@@ -33,31 +55,57 @@ def _parse_args(raw_args):
     parser.add_argument("--version", action="version", version=f"nbQA {__version__}")
     try:
         args, kwargs = parser.parse_known_args(raw_args)
-    except SystemExit as e:
-        if e.code != 0:
+    except SystemExit as exception:
+        if exception.code != 0:
             msg = (
                 "Please specify both a command and a notebook/directory, e.g.\n"
                 "nbqa flake8 my_notebook.ipynb"
             )
-            raise ValueError(msg) from e
+            raise ValueError(msg) from exception
         sys.exit(0)
     command = args.command
     root_dir = args.root_dir
     return command, root_dir, kwargs
 
 
-def _get_notebooks(root_dir) -> List[Path]:
+def _get_notebooks(root_dir: str) -> Iterator[Path]:
     """
     Get generator with all notebooks in directory.
+
+    Parameters
+    ----------
+    root_dir
+        Notebook or directory to run third-party tool on.
+
+    Returns
+    -------
+    notebooks
+        All Jupyter Notebooks found in directory.
     """
     if not Path(root_dir).is_dir():
         return (i for i in (Path(root_dir),))
-    return (i for i in Path(".").rglob("*.ipynb") if ".ipynb_checkpoints" not in str(i))
+    notebooks = (
+        i for i in Path(".").rglob("*.ipynb") if ".ipynb_checkpoints" not in str(i)
+    )
+    return notebooks
 
 
-def _temp_python_file_for_notebook(notebook, tmpdir):
+def _temp_python_file_for_notebook(notebook: Path, tmpdir: str) -> Path:
     """
     Get temporary file to save converted notebook into.
+
+    Parameters
+    ----------
+    notebook
+        Notebook that third-party tool will be run on.
+    tmpdir
+        Temporary directory where converted notebooks will be saved.
+
+    Returns
+    -------
+    Path
+        Temporary Python file whose location mirrors that of the notebook, but
+        inside the temporary directory.
     """
     # Add 3 extra whitespaces because `ipynb` is 3 chars longer than `py`.
     temp_python_file = (
@@ -70,9 +118,29 @@ def _temp_python_file_for_notebook(notebook, tmpdir):
     return temp_python_file
 
 
-def _replace_full_path_out_err(out, err, temp_python_file, notebook):
+def _replace_full_path_out_err(
+    out: str, err: str, temp_python_file: Path, notebook: Path
+) -> Tuple[str, str]:
     """
-    Take care of case when out/err display full path.
+    Replace references to temporary Python file's full path with notebook's path.
+
+    Parameters
+    ----------
+    out
+        Captured stdout from third-party tool.
+    err
+        Captured stderr from third-party tool.
+    temp_python_file
+        Temporary Python file where notebook was converted to.
+    notebook
+        Original Jupyter notebook.
+
+    Returns
+    -------
+    out
+        Stdout with temporary Python file's full path with notebook's path.
+    err
+        Stderr with temporary Python file's full path with notebook's path.
     """
     out = out.replace(str(temp_python_file), str(notebook.resolve()))
     err = err.replace(str(temp_python_file), str(notebook.resolve()))
@@ -85,10 +153,27 @@ def _replace_full_path_out_err(out, err, temp_python_file, notebook):
     return out, err
 
 
-def _replace_relative_path_out_err(out, err, notebook):
+def _replace_relative_path_out_err(
+    out: str, err: str, notebook: Path
+) -> Tuple[str, str]:
     """
-    Take care of case when out/err display relative path.
+    Replace references to temporary Python file's relative path with notebook's path.
 
+    Parameters
+    ----------
+    out
+        Captured stdout from third-party tool.
+    err
+        Captured stderr from third-party tool.
+    notebook
+        Original Jupyter notebook.
+
+    Returns
+    -------
+    out
+        Stdout with temporary Python file's relative path with notebook's path.
+    err
+        Stderr with temporary Python file's relative path with notebook's path.
     Examples
     --------
     >>> out = "notebook   .py ."
@@ -109,22 +194,45 @@ def _replace_relative_path_out_err(out, err, notebook):
     return out, err
 
 
-def _map_python_line_to_nb_lines(out, err, temp_python_file, notebook):
+def _map_python_line_to_nb_lines(
+    out: str, err: str, temp_python_file: Path, notebook: Path
+) -> Tuple[str, str]:
     """
-    Make sure stdout and stderr make reference to Jupyter Notebook lines.
+    Make sure stdout and stderr make reference to Jupyter Notebook cells and lines.
+
+    Parameters
+    ----------
+    out
+        Captured stdout from third-party tool.
+    err
+        Captured stderr from third-party tool.
+    temp_python_file
+        Temporary Python file where notebook was converted to.
+    notebook
+        Original Jupyter notebook.
+
+    Returns
+    -------
+    out
+        Stdout with references to temporary Python file's lines replaced with references
+        to notebook's cells and lines.
+    err
+        Stderr with references to temporary Python file's lines replaced with references
+        to notebook's cells and lines.
     """
     with open(str(temp_python_file), "r") as handle:
         cells = handle.readlines()
     mapping = {}
     cell_no = 0
     cell_count = None
-    for n, i in enumerate(cells):
-        if i == "# %%\n":
+    for idx, cell in enumerate(cells):
+        if cell == "# %%\n":
             cell_no += 1
             cell_count = 0
         else:
+            assert cell_count is not None
             cell_count += 1
-        mapping[n + 1] = f"cell_{cell_no}:{cell_count}"
+        mapping[idx + 1] = f"cell_{cell_no}:{cell_count}"
     out = re.sub(
         rf"(?<={notebook.name}:)\d+", lambda x: str(mapping[int(x.group())]), out,
     )
@@ -135,10 +243,28 @@ def _map_python_line_to_nb_lines(out, err, temp_python_file, notebook):
 
 
 def _replace_temp_python_file_references_in_out_err(
-    temp_python_file, notebook, out, err
-):
+    temp_python_file: Path, notebook: Path, out: str, err: str
+) -> Tuple[str, str]:
     """
-    Replace references to temporary directory name with current working directory.
+    Replace references to temporary Python file with references to notebook.
+
+    Parameters
+    ----------
+    temp_python_file
+        Temporary Python file where notebook was converted to.
+    notebook
+        Original Jupyter notebook.
+    out
+        Captured stdout from third-party tool.
+    err
+        Captured stderr from third-party tool.
+
+    Returns
+    -------
+    out
+        Stdout with temporary directory replaced by current working directory.
+    err
+        Stderr with temporary directory replaced by current working directory.
     """
     out, err = _replace_full_path_out_err(out, err, temp_python_file, notebook)
     out, err = _replace_relative_path_out_err(out, err, notebook)
@@ -146,17 +272,34 @@ def _replace_temp_python_file_references_in_out_err(
     return out, err
 
 
-def _replace_tmpdir_references(out, err, tmpdirname, cwd=None):
-    """
+def _replace_tmpdir_references(
+    out: str, err: str, cwd: Optional[Path] = None
+) -> Tuple[str, str]:
+    r"""
     Replace references to temporary directory name with current working directory.
 
+    Parameters
+    ----------
+    out
+        Captured stdout from third-party tool.
+    err
+        Captured stderr from third-party tool.
+    cwd
+        Current working directory.
+
+    Returns
+    -------
+    out
+        Stdout with references to temporary Python replaced with references to notebook.
+    err
+        Stderr with references to temporary Python replaced with references to notebook.
+
     Examples
     --------
     >>> out = f"rootdir: {os.path.join('tmp', 'tmpdir')}\\n"
     >>> err = ""
-    >>> tmpdirname = os.path.join('tmp', 'tmpdir')
     >>> cwd = Path("nbQA-dev")
-    >>> out, err = _replace_tmpdir_references(out, err, tmpdirname, cwd)
+    >>> out, err = _replace_tmpdir_references(out, err, cwd)
     >>> out.strip(os.linesep)
     'rootdir: nbQA-dev'
     """
@@ -181,9 +324,16 @@ def _replace_tmpdir_references(out, err, tmpdirname, cwd=None):
     return new_out, new_err
 
 
-def _create_blank_init_files(notebook, tmpdirname):
+def _create_blank_init_files(notebook: Path, tmpdirname: str) -> None:
     """
     Replicate local (possibly blank) __init__ files to temporary directory.
+
+    Parameters
+    ----------
+    notebook
+        Notebook third-party tool is being run against.
+    tmpdirname
+        Temporary directory to store converted notebooks in.
     """
     parts = notebook.resolve().relative_to(Path.cwd()).parts
     init_files = Path(parts[0]).rglob("__init__.py")
@@ -192,9 +342,14 @@ def _create_blank_init_files(notebook, tmpdirname):
         Path(tmpdirname).joinpath(i).touch()
 
 
-def _ensure_cell_separators_remain(temp_python_file):
+def _ensure_cell_separators_remain(temp_python_file: Path) -> None:
     """
-    Isort removes a blank line which separates the cells.
+    Reinstate blank line which separates the cells (may be removed by isort).
+
+    Parameters
+    ----------
+    temp_python_file
+        Temporary Python file notebook was converted to.
     """
     with open(str(temp_python_file), "r") as handle:
         py_file = handle.read()
@@ -203,7 +358,9 @@ def _ensure_cell_separators_remain(temp_python_file):
         handle.write(py_file)
 
 
-def _get_arg(root_dir, tmpdirname, nb_to_py_mapping):
+def _get_arg(
+    root_dir: str, tmpdirname: str, nb_to_py_mapping: Dict[Path, Path]
+) -> Path:
     """
     Get argument to run command against.
 
@@ -212,6 +369,20 @@ def _get_arg(root_dir, tmpdirname, nb_to_py_mapping):
     If running against a directory, it'll be the directory mirrored in the temporary
     directory.
 
+    Parameters
+    ----------
+    root_dir
+        Notebook or directory third-party tool is being run against.
+    tmpdirname
+        Temporary directory where converted notebooks are stored.
+    nb_to_py_mapping
+        Mapping between notebooks and Python files corresponding to converted notebooks.
+
+    Returns
+    -------
+    Path
+        Notebook or directory to run third-party tool against.
+
     Examples
     --------
     >>> root_dir = "root_dir"
@@ -230,9 +401,42 @@ def _get_arg(root_dir, tmpdirname, nb_to_py_mapping):
     return arg
 
 
-def _run_command(command, root_dir, tmpdirname, nb_to_py_mapping, kwargs):
+def _run_command(
+    command: str,
+    root_dir: str,
+    tmpdirname: str,
+    nb_to_py_mapping: Dict[Path, Path],
+    kwargs: List[str],
+) -> Tuple[str, str, int]:
     """
     Run third-party tool against given file or directory.
+
+    Parameters
+    ----------
+    command
+        Third-party tool (e.g. :code:`mypy`) to run against notebook.
+    root_dir
+        Notebook or directory to run third-party tool on.
+    tmpdirname
+        Temporary directory where converted notebooks will be stored.
+    nb_to_py_mapping
+        Mapping between notebooks and Python files corresponding to converted notebooks.
+    kwargs
+        Flags to pass to third-party tool (e.g. :code:`--verbose`).
+
+    Returns
+    -------
+    out
+        Captured stdout from running third-party tool.
+    err
+        Captured stderr from running third-party tool.
+    output_code
+        Return code from third-party tool.
+
+    Raises
+    ------
+    ValueError
+        If third-party tool isn't found in system.
     """
     env = os.environ.copy()
     env["PYTHONPATH"] = os.getcwd()
@@ -259,8 +463,16 @@ def _run_command(command, root_dir, tmpdirname, nb_to_py_mapping, kwargs):
     return out, err, output_code
 
 
-def main(raw_args=None):
+def main(raw_args: Optional[List[str]] = None) -> None:
+    """
+    Run third-party tool (e.g. :code:`mypy`) against notebook or directory.
 
+    Parameters
+    ----------
+    raw_args
+        Command-line arguments (if calling this function directly), defaults to
+        :code:`None` if calling via command-line.
+    """
     command, root_dir, kwargs = _parse_args(raw_args)
 
     notebooks = _get_notebooks(root_dir)
@@ -289,7 +501,7 @@ def main(raw_args=None):
             command, root_dir, tmpdirname, nb_to_py_mapping, kwargs
         )
 
-        out, err = _replace_tmpdir_references(out, err, tmpdirname)
+        out, err = _replace_tmpdir_references(out, err)
 
         for notebook, temp_python_file in nb_to_py_mapping.items():
             out, err = _replace_temp_python_file_references_in_out_err(
