@@ -1,9 +1,26 @@
+"""
+Replace :code:`source` code cells of original notebook with ones from converted file.
+
+The converted file will have had the third-party tool run against it by now.
+"""
+
 import json
+from typing import TYPE_CHECKING
+
+if TYPE_CHECKING:  # pragma: nocover
+    from pathlib import Path
 
 
-def main(python_file, notebook):
+def main(python_file: "Path", notebook: "Path") -> None:
     """
-    Replace `source` of original notebook.
+    Replace :code:`source` code cells of original notebook.
+
+    Parameters
+    ----------
+    python_file
+        Temporary Python file notebook was converted to.
+    notebook
+        Jupyter Notebook third-party tool is run against.
     """
     with open(notebook, "r") as handle:
         notebook_json = json.load(handle)
